﻿--[[
    Author: Valana <Team BATTLE> of Mug'thol Alliance-US
]]

local TbMisdirector = LibStub("AceAddon-3.0"):NewAddon("TbMisdirector", "AceComm-3.0", "AceEvent-3.0", "AceSerializer-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("TB_Misdirector", true)
local TbCore = TbCore

local tostring = tostring
local random = random
local pairs = pairs
local time = time
local format = format
local bband = bit.band
local UnitIsUnit = UnitIsUnit
local UnitInParty = UnitInParty
local UnitInRaid = UnitInRaid

--
-- We will process Misdirections affiliated with me or my group.
--
local _player = UnitName("player")
local _raid = bit.bor(COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_AFFILIATION_RAID)
local _announcement = "[%s] %s [%s]"
local _negotiationRank
local _negotiationId
local _masterAnnouncer
local _masterAnnouncerRank
local _misdirectors = { }
local _vigilanceSources = { }


-------------------------------------------------------------------------------
-- Initialization.
-------------------------------------------------------------------------------
function TbMisdirector:OnInitialize()
    --
    -- Our negotiation rank will be properly initialized later when we initiate a brand new negotiation, but just in case we get a
    -- negotiation request in between, initialize it here anyway.
    --
    _negotiationRank = tostring(random())
end

function TbMisdirector:OnEnable()
    --
    -- Update our event registrations.
    --
    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    self:RegisterEvent("PARTY_MEMBERS_CHANGED")

    --
    -- Register for communications received from other Misdirectors.
    --
    self:RegisterComm("TbMisdirector:INITIATE", "INITIATE")
    self:RegisterComm("TbMisdirector:NEGOTIATE", "NEGOTIATE")
    self:RegisterComm("TbMisdirector:NEGOTIATION_RESPONSE", "NEGOTIATION_RESPONSE")

    --
    -- Start a negotiation in a bit if we loaded into a group.
    --
    self:ScheduleTimer(function() if (TbCore:IsInGroup()) then self:StartNegotiation(true) end end, 5)
end


-------------------------------------------------------------------------------
-- Event handlers.
-------------------------------------------------------------------------------
function TbMisdirector:PARTY_MEMBERS_CHANGED(e)
    if (TbCore:IsInGroup()) then
        if (_masterAnnouncer) then
            --
            -- Verify that all members running Misdirector are still around.
            --
            local checker = UnitInParty
            if (TbCore:IsInRaid()) then checker = UnitInRaid end
            for misdirector, online in pairs(_misdirectors) do
                if (not checker(misdirector)) then
                    _misdirectors[misdirector] = nil
                end
            end

            --
            -- If we have just lost our master announcer, re-negotiate on our own because all other Misdirectors will be`
            -- doing this independently as well.
            --
            if (not _misdirectors[_masterAnnouncer] and not UnitIsUnit("player", _masterAnnouncer)) then
                self:StartNegotiation()
            end
        else
            --
            -- We have just joined a group. Tell everybody to re-negotiate.
            --
            self:StartNegotiation(true)
        end
    else
        --
        -- We have just left our group. Forget about all misdirectors. The current group members will figure out whether they need to re-negotiate.
        --
        _masterAnnouncer = nil
        wipe(_misdirectors)
    end
end

--
-- Process the combat log to announce Misdirections.
--
function TbMisdirector:COMBAT_LOG_EVENT_UNFILTERED(e, timestamp, event, sourceGuid, sourceName, sourceFlags, destinationGuid, destinationName, destinationFlags, spellId, spellName, ...)
    --
    -- Filter out events we're not interested in.
    --
    local group = TbCore:GetRaidOrParty()
    if (not group or bband(sourceFlags, _raid) == 0 or event ~= "SPELL_CAST_SUCCESS") then return end

    --
    -- If the spell cast was Vigilance, we need to keep track of who is casting it so we may may blacklist spammers. Another solution would be to synchronize
    -- lists among misdirectors but that results in unnecessary network traffic.
    --
    if (spellId == 50720) then
        if (not _vigilanceSources[sourceName]) then
            --
            -- This is the first time this warrior has cast Vigilance this session. Initialize his or her data.
            --
            self:ResetWarrior(sourceName)
        else
            local t = time()
            local elapsed = t - _vigilanceSources[sourceName]["time"]
            if (_vigilanceSources[sourceName]["count"] < 3) then
                --
                -- Block excessive Vigilance casts within 15 seconds of each other.
                --
                if (elapsed > 15) then
                    self:ResetWarrior(sourceName)
                else
                    _vigilanceSources[sourceName]["time"] = t
                    _vigilanceSources[sourceName]["count"] = _vigilanceSources[sourceName]["count"] + 1

                    --
                    -- If this is the warrior's third Vigilance cast, put him or her on probation, instead of blacklisting him or
                    -- her outright.
                    --
                    if (_vigilanceSources[sourceName]["count"] == 3) then return end
                end
            --
            -- Ignore this warrior's Vigilance casts for a minute because he or she has been spamming.
            --
            elseif (elapsed < 60) then return
            else
                self:ResetWarrior(sourceName)
            end
        end
        if (_masterAnnouncer and UnitIsUnit("player", _masterAnnouncer)) then
            SendChatMessage(format(_announcement, sourceName, L["ANNOUNCE_VIGILANCE"], destinationName), group)
        end
    elseif (not _masterAnnouncer or not UnitIsUnit("player", _masterAnnouncer)) then return end

    if (spellId == 34477) then
        SendChatMessage(format(_announcement, sourceName, L["ANNOUNCE_MISDIRECTION"], destinationName), group)
    elseif (spellId == 57934) then
        SendChatMessage(format(_announcement, sourceName, L["ANNOUNCE_TRICKS_OF_THE_TRADE"], destinationName), group)
    end
end

--
-- React to communications received from group members.
--
function TbMisdirector:INITIATE(prefix, message, distribution, sender)
    if (self:ShouldIgnoreMessage(message, sender)) then return end
    self:StartNegotiation()
end

function TbMisdirector:NEGOTIATE(prefix, negotiationId, distribution, sender)
    if (self:ShouldIgnoreMessage(message, sender)) then return end
    self:SendCommMessage("TbMisdirector:NEGOTIATION_RESPONSE", self:Serialize(negotiationId, _negotiationRank), "WHISPER", sender)
end

function TbMisdirector:NEGOTIATION_RESPONSE(prefix, message, distribution, sender)
    if (self:ShouldIgnoreMessage(message, sender)) then return end

    local success, negotiationId, negotiationRank = self:Deserialize(message)
    if (success and negotiationId == _negotiationId) then
        _misdirectors[sender] = true
        if (negotiationRank > _masterAnnouncerRank or (negotiationRank == _masterAnnouncerRank and sender > _masterAnnouncer)) then
            _masterAnnouncer, _masterAnnouncerRank = sender, negotiationRank
        end
    end
end


-------------------------------------------------------------------------------
-- Helper methods.
-------------------------------------------------------------------------------
function TbMisdirector:ResetWarrior(name)
    if (not _vigilanceSources[name]) then
        _vigilanceSources[name] = { }
    end
    _vigilanceSources[name]["time"] = time()
    _vigilanceSources[name]["count"] = 1
end

function TbMisdirector:StartNegotiation(askInitiate)
    --
    -- Re-randomize our negotiation rank if we are starting a brand new negotiation. Higher ranks have higher precedence
    -- in becoming master announcers.
    --
    if (askInitiate) then
        _negotiationRank = tostring(random())
    end

    --
    -- Initially, assume we will be the master announcer.
    --
    _masterAnnouncer, _masterAnnouncerRank = _player, _negotiationRank

    --
    -- Tell everyone to re-negotiate the master announcer.
    --
    local group = TbCore:GetRaidOrParty()
    if (askInitiate) then
        self:SendCommMessage("TbMisdirector:INITIATE", "", group, nil)
    end

    --
    -- Generate a random number which will uniquely identify this negotiation so that we may recognize incoming responses.
    --
    _negotiationId = tostring(random())
    self:SendCommMessage("TbMisdirector:NEGOTIATE", _negotiationId, group, nil)
end

function TbMisdirector:ShouldIgnoreMessage(message, sender)
    return UnitIsUnit("player", sender) or not TbCore:IsUnitInGroup(sender) or TbCore:IsIgnored(sender)
end
